#
# $Id:$
#
# This file is part of the Cloud Services Integration Platform (CSIP),
# a Model-as-a-Service framework, API, and application suite.
#
# 2012-2020, OMSLab, Colorado State University.
#
# OMSLab licenses this file to you under the MIT license.
# See the LICENSE file in the project root for more information.
#
from builtins import Exception

import requests
import numpy as np
import json
from csip import Client

NORM = {
    'kge': lambda x: 1 - x,
    'absdiff': lambda x: x,
    'kge09': lambda x: 1 - x,
    'nslog': lambda x: 1 - x,
    'nslog1p': lambda x: 1 - x,
    'nslog2': lambda x: 1 - x,  # the good one???
    'ns': lambda x: 1 - x,
    'mns': lambda x: 1 - x,
    'rmse': lambda x: x,
    'trmse': lambda x: x,
    'pbias': lambda x: abs(x)  # correct one
    # 'pbias': lambda x: abs(2.5 - x)
}


def calc_cost(response: "Client", objfunc) -> float:
    """aggregated objective function value -> cost"""

    # this will eliminate the failed run as a candidate solution.
    if not response.is_finished():
        return np.nan

    cost = 0.0
    for o in objfunc:
        cosu_val = float(response.get_cosu_value(o['name']))
        if cosu_val is None:
            raise Exception('No cosu of value for ' + o['name'])
        cosu_of = o['of']
        n = NORM.get(cosu_of, lambda x: x)  # default
        cost += n(cosu_val) * float(o.get('weight', 1.0))  # original
        if cost < 0:
            print(o['name'], ' ', cosu_of, ' ', cosu_val)
    return cost

def save(file, dict):
    """ Save dict to json file. """
    with open(file, 'w') as json_file:
        json.dump(dict, json_file, sort_keys=False, indent=2)


def load(file):
    """ Load dict from json file. """
    with open(file) as json_file:
        data = json.load(json_file)
    return data


def get_step_info(steps, index):
    """Extract all relevant info from the step dict"""

    step = steps[index]
    l = len(step['param'])
    param_min = []
    param_max = []
    param_names = []

    # extract names and bounds
    for i, p in enumerate(step['param']):
        value_type = p.get('type', 'float')
        if 'float' == value_type:
            param_names.append(p['name'])
            if len(p['bounds']) != 2:
                raise Exception('Invalid bounds tuple: (min, max): "{}"'.format(p['bounds']))
            if not p['bounds'][0] < p['bounds'][1]:
                raise Exception('Invalid bounds values: "{}"'.format(p['bounds']))
            param_min.append(p['bounds'][0])
            param_max.append(p['bounds'][1])
        elif 'list' == value_type:
            if 'default_value' not in p:
                raise Exception('List value types require a default value')
            if type([]) != type(p['default_value']):
                raise Exception('Default value must be a list')

            base_name = p['name']
            calibration_strategy = p.get('calibration_strategy', 'mean')
            if 'mean' == calibration_strategy:
                param_names.append(base_name)
                if len(p['bounds']) != 2:
                    raise Exception('Invalid bounds tuple: (min, max): "{}"'.format(p['bounds']))
                if not p['bounds'][0] < p['bounds'][1]:
                    raise Exception('Invalid bounds values: "{}"'.format(p['bounds']))
                param_min.append(p['bounds'][0])
                param_max.append(p['bounds'][1])
            elif 'single' == calibration_strategy:
                if len(p['bounds']) != 2:
                    raise Exception('Invalid bounds tuple: (min, max): "{}"'.format(p['bounds']))
                if not p['bounds'][0] < p['bounds'][1]:
                    raise Exception('Invalid bounds values: "{}"'.format(p['bounds']))
                for index, value in enumerate(p['default_value']):
                    param_names.append("{}_{}".format(base_name, index))
                    param_min.append(p['bounds'][0])
                    param_max.append(p['bounds'][1])
        else:
            raise Exception('Invalid value type: {}'.format(value_type))

        # check if OF is supported
    for o in step['objfunc']:
        found = False
        for name in NORM:
            if str(o['name']).startswith(name):
                found = True
        if not found:
            raise Exception('OF not supported: "{}"'.format(o['name']))
        if len(o['data']) != 2:
            raise Exception('OF missing data: (sim, obs): "{}"'.format(o['name']))

    return param_names, (param_min, param_max), step['objfunc']


def get_calibrated_params(steps, index):
    """Get all previously calibrated parameter from any other step"""

    step = steps[index]
    cp = {}
    for s in steps:
        # skip the own step
        if s is step:
            continue
        for p in s['param']:
            # if 'value' in p.keys():
            if 'value' in p:
                cp[p['name']] = p['value']
            elif 'default_value' in p:
                cp[p['name']] = p['default_value']

    # if the step parameter are in any other step, take them out since
    # we rather want to calibrate them
    for p in step['param']:
        if p['name'] in cp:
            cp.pop(p['name'])

    return cp


def annotate_step(best_cost, pos, steps, index):
    """Annotate the step with the best value"""

    step = steps[index]
    step['cost'] = best_cost
    value_index = 0
    for i, p in enumerate(step['param']):
        value_type = p.get('type', 'float')
        if 'float' == value_type:
            p['value'] = pos[value_index]
            value_index += 1
        elif 'list' == value_type:
            calibration_strategy = p.get('calibration_strategy', 'mean')
            if 'mean' == calibration_strategy:
                mean_value = pos[value_index]
                p['mean_value'] = mean_value
                p['value'] = [element + mean_value * element for element in p['default_value']]
                value_index += 1
            elif 'single' == calibration_strategy:
                val = []
                j = 0
                while j < len(p['default_value']):
                    val.append(pos[value_index])
                    value_index += 1
                    j += 1
                p['value'] = val


def check_url(url):
    """Check is the Url is valid."""

    r = requests.head(url)
    if r.status_code != 200:
        raise Exception('Error code {} from Url: "{}"'.format(r.status_code, url))



from customtkinter import CTkScrollableFrame
from customtkinter import CTkFrame
from customtkinter import CTkLabel
from customtkinter import CTkButton
from customtkinter import CTkEntry
from customtkinter import CTkOptionMenu
from tkinter import StringVar as sv
import tkinter as tk

class ParameterView(CTkScrollableFrame):
    def __init__(self, *args,
                 option_manager: None,
                 list_name: None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        
        self.option_manager = option_manager
        self.list_name = list_name
        self.key_values = option_manager.get_list(self.list_name)
        self.edit_mode = False
        
        self.render()

    def clear(self):
        self.key_values = self.option_manager.get_list(self.list_name)
        self.containerFrame.destroy()
        
    def toggle_edit_mode(self):
        self.clear()
        self.edit_mode = not self.edit_mode
        self.render()
        
    def render(self):
        row = 0
        index = 0
        
        self.containerFrame = CTkFrame(self)
        self.containerFrame.grid(row=0, column=0, padx=(5, 5), pady=(5, 5), sticky="nsew")
        self.containerFrame.grid_columnconfigure((0, 1, 2, 3, 4, 5), weight=1)

        #CTkLabel(self.containerFrame, text="Name:").grid(row=row, column=0, columnspan=3, padx=5, pady=5, sticky="")
        #CTkLabel(self.containerFrame, text="Value:").grid(row=row, column=3, columnspan=3, padx=5, pady=5, sticky="")
        #row += 1
        
        for key_value_pair in self.key_values:
            required = False
            visual_name = "NONE"
            if "required" in key_value_pair:
                required = key_value_pair["required"].get()
            if required and "visual_name" in key_value_pair:
                visual_name = key_value_pair["visual_name"]

            if visual_name != "NONE":
                CTkLabel(self.containerFrame, text=visual_name.get()).grid(row=row, column=0, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="ew")
                #CTkEntry(self.containerFrame, state="disabled", textvariable=self.key_values[index]["visual_name"]).grid(row=row, column=0, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="ew")
            else:
                CTkEntry(self.containerFrame, textvariable=self.key_values[index]["name"]).grid(row=row, column=0, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="ew")
            
            if self.edit_mode and not required:
                type_menu = CTkOptionMenu(self.containerFrame, variable=self.key_values[index]["type"], values=["integer", "float", "date", "boolean", "string"], width=20)
                type_menu.grid(row=row, column=3, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")

                dest_menu = CTkOptionMenu(self.containerFrame, variable=self.key_values[index]["destination"], values=["args", "kwargs", "conf", "options", "oh_strategy", "metainfo"], width=20)
                dest_menu.grid(row=row, column=4, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")


                return_func = lambda index=index: (self.clear(), self.option_manager.remove_key_value(self.list_name, index), self.render())
                
                CTkButton(self.containerFrame, text="Remove", command=return_func).grid(row=row, column=5, columnspan=1, padx=(5, 5), pady=(5, 5), sticky="ew")
            else:
                type = self.key_values[index]["type"].get()

                if type == "boolean":
                    bb = CTkOptionMenu(self.containerFrame, values=["True", "False", "NULL"], variable=self.key_values[index]["value"])
                    bb.grid(row=row, column=3, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="ew")
                elif type == "datetime":
                    om_text_var = self.key_values[index]["value"]
                    year = sv()
                    month = sv()
                    day = sv()
                    hour = sv()
                    minute = sv()
                    second = sv()

                    year.set("2021")
                    month.set("01")
                    day.set("01")
                    hour.set("00")
                    minute.set("00")
                    second.set("00")

                    try:
                        date_time = om_text_var.get().split(" ")
                        vv = date_time[0].split("-")
                        ss = date_time[1].split(":")
                        year.set(vv[0])
                        month.set(vv[1])
                        day.set(vv[2])
                        hour.set(ss[0])
                        minute.set(ss[1])
                        second.set(ss[2])
                    except Exception as e:
                        pass
                    year_options = [str(i) for i in range(1999, 2030)]
                    month_options = ["01", "02", "03", "04", "05", "06", "07", "08", "09", "10", "11", "12"]
                    day_options = ["01", "02", "03", "04", "05", "06", "07", "08", "09", "10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "20", "21", "22", "23", "24", "25", "26", "27", "28", "29", "30", "31"]
                    hour_options = [str(i) for i in range(0, 24)]
                    minute_options = [f"{i:02}" for i in range(0, 60)]
                    second_options = [f"{i:02}" for i in range(0, 60)]

                    update_date = lambda om_text_var=om_text_var, year=year, month=month, day=day: om_text_var.set(f"{year.get()}-{month.get()}-{day.get()} {hour.get()}:{minute.get()}:{second.get()}")

                    datetime_module = CTkFrame(self.containerFrame)
                    datetime_module.grid(row=row, column=3, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="ew")

                    year_menu = CTkOptionMenu(datetime_module, variable=year, values=year_options, width=50, command=update_date)
                    year_menu.grid(row=0, column=0, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")
                    month_menu = CTkOptionMenu(datetime_module, variable=month, values=month_options, width=20, command=update_date)
                    month_menu.grid(row=0, column=1, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")
                    day_menu = CTkOptionMenu(datetime_module, variable=day, values=day_options, width=20, command=update_date)
                    day_menu.grid(row=0, column=2, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")
                    hour_menu = CTkOptionMenu(datetime_module, variable=hour, values=hour_options, width=20, command=update_date)
                    hour_menu.grid(row=1, column=0, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")
                    minute_menu = CTkOptionMenu(datetime_module, variable=minute, values=minute_options, width=20, command=update_date)
                    minute_menu.grid(row=1, column=1, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")
                    second_menu = CTkOptionMenu(datetime_module, variable=second, values=second_options, width=20, command=update_date)
                    second_menu.grid(row=1, column=2, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")

                elif type == "date":
                    om_text_var = self.key_values[index]["value"]
                    year = sv()
                    month = sv()
                    day = sv()
                    year.set("2021")
                    month.set("01")
                    day.set("01")

                    try:
                        vv = om_text_var.get().split("-")
                        year.set(vv[0])
                        month.set(vv[1])
                        day.set(vv[2])
                    except Exception as e:
                        pass
                    year_options = [str(i) for i in range(1999, 2030)]
                    month_options = ["01", "02", "03", "04", "05", "06", "07", "08", "09", "10", "11", "12"]
                    day_options = ["01", "02", "03", "04", "05", "06", "07", "08", "09", "10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "20", "21", "22", "23", "24", "25", "26", "27", "28", "29", "30", "31"]
                    
                    update_date = lambda om_text_var=om_text_var, year=year, month=month, day=day: om_text_var.set(f"{year.get()}-{month.get()}-{day.get()}")
                    year_menu = CTkOptionMenu(self.containerFrame, variable=year, values=year_options, width=50, command=update_date)
                    year_menu.grid(row=row, column=3, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")
                    month_menu = CTkOptionMenu(self.containerFrame, variable=month, values=month_options, width=20, command=update_date)
                    month_menu.grid(row=row, column=4, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")
                    day_menu = CTkOptionMenu(self.containerFrame, variable=day, values=day_options, width=20, command=update_date)
                    day_menu.grid(row=row, column=5, columnspan=1, padx=(0, 0), pady=(5, 5), sticky="ew")
                else:
                    bb = CTkEntry(self.containerFrame)
                    bb.grid(row=row, column=3, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="ew")
                    bb.configure(textvariable=self.key_values[index]["value"])
            row += 1
            index += 1
            
        if self.edit_mode:
            CTkButton(self.containerFrame, text="Exit", command=self.toggle_edit_mode).grid(row=row, column=0, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="ew")
        else:
            CTkButton(self.containerFrame, text="Edit", command=self.toggle_edit_mode).grid(row=row, column=0, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="ew")
            
        add_key_func = lambda: (self.clear(), self.option_manager.add_key_value(self.list_name, "name", "value"), self.render())

        CTkButton(self.containerFrame, text="Add Parameter", command=add_key_func).grid(row=row, column=3, columnspan=3, padx=(5, 5), pady=(5, 5), sticky="ew")


from tkinter import StringVar as sv
import json
import os

class OptionManager():
    
    def __init__(self):

        self._service_modes = ["Sampling: Halton", "Sampling: Random", "Sensitivity Analysis", "Optimization"]

        self._default_sampling = json.load(open(os.path.join("./defaults", "sampling.json")))
        self._default_sensitivity = json.load(open(os.path.join("./defaults", "sensitivity.json")))
        self._default_optimization = json.load(open(os.path.join("./defaults", "optimization.json")))

        self._mode_sv = sv()
        self.init_lists()

    def init_lists(self): 
        self._project_data = {"name": "cosu_default", "path": "/tmp"}
        self._data = {}

        self._mode_sv.set("Sampling: Halton")

        for service in self._service_modes:
            self._data[service] = {
                    "url": sv(),
                    "files": {},
                    "steps": [],
                    "model_parameters": [],
                    "hyperparameters": [],
                    "service_parameters": [],
                    "service_request_data": [],
                    "figure_parameters": [],
                    "sensitivity_parameters": [],
                    "sensitivity_positiveBestMetrics": [],
                    "sensitivity_negativeBestMetrics": [],

                    "sensitivity_analysis_path": sv(),
                    "optimization_selected_group": sv(),
                    "sampling_output_mode": sv(),
                    "selected_graph": sv(),
                    "graph_theme": sv(),
                    "selected_csv": sv(),
                    "selected_csv2": sv(),
                    "selected_x": sv(),
                    "selected_y1": sv(),
                    "selected_y2": sv(),
                    "figure_style": sv(),
                    "matrix_values": []
                }
            
            self._data[service]["sensitivity_analysis_path"].set("No file selected...")
            self._data[service]["sampling_output_mode"].set("Replace")
            self._data[service]["selected_graph"].set("None")
            self._data[service]["graph_theme"].set("Dark")
            self._data[service]["selected_csv"].set("No files found...")
            self._data[service]["selected_csv2"].set("No files found...")
            self._data[service]["selected_x"].set("time")
            self._data[service]["selected_y1"].set("NONE")
            self._data[service]["selected_y2"].set("NONE")
            self._data[service]["figure_style"].set("Scatter")
            self._data[service]["matrix_values"].append(sv())
            self._data[service]["matrix_values"][0].set("NONE")

            
            if service == "Sampling: Halton" or service == "Sampling: Random":
                self._data[service]["model_parameters"] = self.deserialize_data(self._default_sampling["model_parameters"])
                self._data[service]["hyperparameters"] = self.deserialize_data(self._default_sampling["hyperparameters"])
                self._data[service]["service_parameters"] = self.deserialize_data(self._default_sampling["service_parameters"])
            elif service == "Sensitivity Analysis":
                self._data[service]["model_parameters"] = self.deserialize_data(self._default_sensitivity["model_parameters"])
                self._data[service]["hyperparameters"] = self.deserialize_data(self._default_sensitivity["hyperparameters"])
                self._data[service]["service_parameters"] = self.deserialize_data(self._default_sensitivity["service_parameters"])
            elif service == "Optimization":
                self._data[service]["model_parameters"] = self.deserialize_data(self._default_optimization["model_parameters"])
                self._data[service]["hyperparameters"] = self.deserialize_data(self._default_optimization["hyperparameters"])
                self._data[service]["service_parameters"] = self.deserialize_data(self._default_optimization["service_parameters"])

            
    def serialize_data(self, data):
        if isinstance(data, dict):
            return {key: self.serialize_data(value) for key, value in data.items()}
        elif isinstance(data, list):
            return [self.serialize_data(item) for item in data]
        elif isinstance(data, sv):
            return data.get()
        else:
            return data
        
    def deserialize_data(self, data):
        if isinstance(data, dict):
            return {key: self.deserialize_data(value) for key, value in data.items()}
        elif isinstance(data, list):
            return [self.deserialize_data(item) for item in data]
        elif isinstance(data, (str, int, float)):
            return sv(value=str(data))
        else:
            return data
        
    def save_project(self, filename):
        with open(filename, 'w') as file:
            results = {"project_data": self._project_data, 
                       "mode": self._mode_sv.get(),
                       "data": self.serialize_data(self._data)}
            json.dump(results, file)

    def get_all_data(self):
        return self.serialize_data(self._data)

    def load_project(self, filename):
        
        print("Initizializing lists...")
        self.init_lists()

        with open(filename, 'r') as file:
            results = json.load(file)
            new_project_data = results["project_data"]
            for key, value in new_project_data.items():
                self._project_data[key] = value
            self._mode_sv.set(results["mode"])

            print("Deserializing data...")
            new_data = self.deserialize_data(results["data"])

            for service in self._service_modes:
                print("Processing data for " + service + "...")
                for key, value in new_data[service].items():
                    self._data[service][key] = value


    def add_arguments(self, arguments):
            
        if ("mode" in arguments):
            self._data["mode"].set(arguments["mode"])

        if ("url" in arguments):
            self._data["url"][self._data["mode"].get()].set(arguments["url"])
            
        if ("files" in arguments):
            for file in arguments["files"]:
                name = file["name"]
                value = file["value"]
                obj = {"name": sv(), "value": sv()}
                obj["name"].set(name)
                obj["value"].set(value)
                self._data["files"][name] = obj
        
        for param in arguments["model_parameters"]:
            name = param["name"]
            value = param["value"]
            obj = {"name": sv(), "value": sv()}
            obj["name"].set(name)
            obj["value"].set(value)
            self._data["model_parameters"].append(obj)
            
        for param in arguments["hyperparameters"]:
            name = param["name"]
            value = param["value"]
            obj = {"name": sv(), "value": sv()}
            obj["name"].set(name)
            obj["value"].set(value)
            self._data["hyperparameters"].append(obj)

        for param in arguments["service_parameters"]:
            name = param["name"]
            value = param["value"]
            obj = {"name": sv(), "value": sv()}
            obj["name"].set(name)
            obj["value"].set(value)
            self._data["service_parameters"].append(obj)
            
    def add_steps(self, steps):
        for step in steps:
            obj = {"parameter_objects": [], 
                    "override_parameter": [],
                    "objective_functions": [], 
                    "name": sv(), 
                    "open": True}
            obj["name"].set("Group " + str(len(self._data[self._mode_sv.get()]["steps"]) + 1))
            
            if "parameter_objects" in step:
                for param in step["parameter_objects"]:
                    param_obj = {
                        "name": sv(), 
                        "min_bound": sv(),
                        "max_bound": sv(),
                        "default_value": sv(), 
                        "optimal_value": sv(), 
                        "type": sv(),
                        "calibration_strategy": sv()
                    }
                    param_obj["name"].set(param["name"])

                    if "min_bound" in param:
                        param_obj["min_bound"].set(param["min_bound"]) 
                    else: 
                        param_obj["min_bound"].set(0)

                    if "max_bound" in param:
                        param_obj["max_bound"].set(param["max_bound"]) 
                    else: 
                        param_obj["max_bound"].set(0)
                        
                    if "type" in param:
                        param_obj["type"].set(param["type"])
                    else:
                        param_obj["type"].set("float")
                        
                    if "default_value" in param:
                        param_obj["default_value"].set(param["default_value"])
                    else:
                        param_obj["default_value"].set(1)
                        
                    if "optimal_value" in param:
                        param_obj["optimal_value"].set(param["optimal_value"])
                    else:
                        param_obj["optimal_value"].set(0)
                        
                    if "calibration_strategy" in param:
                        param_obj["calibration_strategy"].set(param["calibration_strategy"])
                    else:
                        param_obj["calibration_strategy"].set("none")
                        
                    obj["parameter_objects"].append(param_obj)
            
            if "override_parameter" in step:
                for override in step["override_parameter"]:
                    override_obj = {"name": sv(), "value": sv()}
                    override_obj['name'].set(override['name'])
                    override_obj['value'].set(override['value'])
                    obj['override_parameter'].append(override_obj)

            if "objective_functions" in step:
                for objective_function in step["objective_functions"]:
                    objective_function_object = {"name": sv(), 
                                    "objective_function": sv(), 
                                    "weight": sv(), 
                                    "custom_function": sv(),
                                    "custom_function_goal": sv(),
                                    "custom_function_value": sv(),
                                    "data_observed": sv(),
                                    "data_simulated": sv()}
                    objective_function_object["name"].set(objective_function["name"])
                    objective_function_object["objective_function"].set(objective_function["objective_function"])
                    objective_function_object["custom_function_goal"].set("Positive Best")
                    
                    if ("weight" in objective_function): 
                        objective_function_object["weight"].set(objective_function["weight"])
                    else:
                        objective_function_object["weight"].set(1)
                        
                    if ("custom_function" in objective_function):
                        objective_function_object["custom_function"].set(objective_function["custom_function"])
                    if ("custom_function_goal" in objective_function):
                        objective_function_object["custom_function_goal"].set(objective_function["custom_function_goal"])
                    if ("custom_function_value" in objective_function):
                        objective_function_object["custom_function_value"].set(objective_function["custom_function_value"])

                    objective_function_object["data_observed"].set(objective_function["data_observed"])
                    objective_function_object["data_simulated"].set(objective_function["data_simulated"]) 
                    obj["objective_functions"].append(objective_function_object)
            
            self._data[self._mode_sv.get()]["steps"].append(obj)
    
    def add_function(self, step_index):
        obj = {"name": sv(), 
                "objective_function": sv(), 
                "weight": sv(), 
                "custom_function": sv(),
                "data_observed": sv(),
                "data_simulated": sv()}
        obj["name"].set("ns")
        obj["objective_function"].set("ns")
        obj["weight"].set(1)
        obj["data_observed"].set("")
        obj["data_simulated"].set("")
        obj["custom_function"].set("") 
        
        self._data[self._mode_sv.get()]["steps"][step_index]["objective_functions"].append(obj)
        
    def remove_function(self, step_index, index):
        self._data[self._mode_sv.get()]["steps"][step_index]["objective_functions"].pop(index)
        
    def dupe_function(self, step_index, index):
        my_func = self._data[self._mode_sv.get()]["steps"][step_index]["objective_functions"][index]
        
        new_object = {"name": sv(), 
                        "objective_function": sv(), 
                        "weight": sv(), 
                        "custom_function": sv(),
                        "data_observed": sv(),
                        "data_simulated": sv()}
        new_object["name"].set(my_func["name"].get())
        new_object["objective_function"].set(my_func["objective_function"].get())
        new_object["weight"].set(my_func["weight"].get())
        new_object["data_observed"].set(my_func["data_observed"].get())
        new_object["data_simulated"].set(my_func["data_simulated"].get())
        new_object["custom_function"].set(my_func["custom_function"].get())
        
        self._data[self._mode_sv.get()]["steps"][step_index]["objective_functions"].append(new_object)
    
    def add_bound(self, step_index, 
                  name="name", 
                  min=0, 
                  max=1,
                  type="float",
                  default_value=1,
                  optimal_value=0,
                  calibration_strategy="none"):
        obj = {
            "name": sv(), 
            "min_bound": sv(),
            "max_bound": sv(),
            "default_value": sv(), 
            "optimal_value": sv(),
            "type": sv(),
            "calibration_strategy": sv()
        }
        obj["name"].set(name)
        obj["type"].set(type)
        obj["default_value"].set(default_value)
        obj["optimal_value"].set(optimal_value)
        obj["calibration_strategy"].set(calibration_strategy)
        obj["min_bound"].set(min)
        obj["max_bound"].set(max)
        self._data[self._mode_sv.get()]["steps"][step_index]["parameter_objects"].append(obj)
        
    def remove_bound(self, step_index, index):
        self._data[self._mode_sv.get()]["steps"][step_index]["parameter_objects"].pop(index)
        
    def add_override(self, step_index, name, value):
        obj = {"name": sv(), "value": sv()}
        obj["name"].set(name)
        obj["value"].set(value)
        self._data[self._mode_sv.get()]["steps"][step_index]["override_parameter"].append(obj)

    def remove_override(self, step_index, index):
        self._data[self._mode_sv.get()]["steps"][step_index]["override_parameter"].pop(index)

    def get_override(self, step_index):
        return self._data[self._mode_sv.get()]["steps"][step_index]["override_parameter"]

    def add_key_value(self, list_name, key, value, type="string", destination="args"):
        obj = {"name": sv(), "value": sv(), "type": sv(), "destination": sv()}
        obj["name"].set(key)
        obj["value"].set(value)
        obj["type"].set(type)
        obj["destination"].set(destination)
        self._data[self._mode_sv.get()][list_name].append(obj)

    def remove_key_value(self, list_name, index):
        self._data[self._mode_sv.get()][list_name].pop(index)
            
    def move_step_up(self, index):
        if index > 0:
            self._data[self._mode_sv.get()]["steps"][index], self._data[self._mode_sv.get()]["steps"][index - 1] = self._data[self._mode_sv.get()]["steps"][index - 1], self._data[self._mode_sv.get()]["steps"][index]
            
    def move_step_down(self, index):
        if index < len(self._data[self._mode_sv.get()]["steps"]) - 1:
            self._data[self._mode_sv.get()]["steps"][index], self._data[self._mode_sv.get()]["steps"][index + 1] = self._data[self._mode_sv.get()]["steps"][index + 1], self._data[self._mode_sv.get()]["steps"][index]
            
    def toggle_step_open(self, index):
        self._data[self._mode_sv.get()]["steps"][index]["open"] = not self._data[self._mode_sv.get()]["steps"][index]["open"]

    def remove_step(self, index):
        self._data[self._mode_sv.get()]["steps"].pop(index)
            
    def get_project_data(self):
        return self._project_data
            
    def set_path(self, filename):
        file_name = filename.split("/")[-1].replace(".json", "")
        path = filename.replace(file_name + ".json", "")
        self._project_data["path"] = path
        self._project_data["name"] = file_name

    def copy_list(self, source_mode):
        name = "backup"
        index = 0

        target_mode = self._mode_sv.get()

        while name + str(index) in self._data:
            index += 1
        self._data[name + str(index)] = self._data[target_mode]

        # Copy data except hyperparameters
        original_hyperparameters = self._data[target_mode]["hyperparameters"]
        self._data[target_mode] = self._data[source_mode]
        self._data[target_mode]["hyperparameters"] = original_hyperparameters

        # Add in cal_startTime and cal_endTime if missing, copy values from startTime/endTime
        if target_mode == "Optimization":
            startTime = None
            endTime = None
            cal_startTime = None
            cal_endTime = None
            for param in self._data[target_mode]["model_parameters"]:
                if param["name"] == "startTime":
                    startTime = param
                elif param["name"] == "endTime":
                    endTime = param
                elif param["name"] == "cal_startTime":
                    cal_startTime = param
                elif param["name"] == "cal_endTime":
                    cal_endTime = param

            if cal_endTime is None and startTime is not None:
                startTime["name"] = "cal_startTime"
                self._data[target_mode]["model_parameters"].append(startTime)

            if cal_startTime is None and endTime is not None:
                endTime["name"] = "cal_endTime"
                self._data[target_mode]["model_parameters"].append(endTime)

    def combine_steps(self):
        mode = self._mode_sv.get()

        obj = {"parameter_objects": [], 
            "override_parameter": [],
            "objective_functions": [], 
            "name": sv(), 
            "open": True}
        obj["name"].set("Combined Group")

        for step in self._data[mode]["steps"]:
            for param in step["parameter_objects"]:
                obj["parameter_objects"].append(param)
            for objective_function in step["objective_functions"]:
                obj["objective_functions"].append(objective_function)
            for override in step["override_parameter"]:
                obj["override_parameter"].append(override)
            
        self._data[mode]["steps"] = [obj]
            
    def get_data(self):
        return self._data[self._mode_sv.get()]
    
    def get_steps(self):
        return self._data[self._mode_sv.get()]["steps"]
    
    def get_mode(self):
        return self._mode_sv.get()
    
    def get_mode_sv(self):
        return self._mode_sv
    
    def get_service_modes(self):
        return self._service_modes
    
    def get_list(self, list_name):
        return self._data[self._mode_sv.get()][list_name]
    
    def get(self, key):
        return self._data[self._mode_sv.get()][key]
    
    def set_data(self, key, value):
        self._data[self._mode_sv.get()][key] = value

    def set_var(self, key, value):
        self._data[self._mode_sv.get()][key].set(value)
        
    def get_all_as_json(self):
        obj = {"arguments": self._data, "steps": self._data[self._mode_sv.get()]["steps"]}
        return obj
    
    def get_project_folder(self):
        return os.path.join(self._project_data['path'], self._project_data['name'])

    
    def get_metrics(self):

        self._data["url"].set(self._data["urls"][self._data["mode"].get()].get())

        result = {}
        result['arguments'] = {}
        result['hyperparameters'] = []
        result['service_parameters'] = []
        result['service_parameters'] = {}
        result['project_data'] = self._project_data
        for key, value in self._data.items():
            if key == 'url' or key == 'mode':
                result['arguments'][key] = value.get()
            elif key == 'files':
                result['arguments'][key] = {}
                #for name, obj in value.items():
                #    result['arguments'][key].append({'name': obj['name'].get(), 'value': obj['value'].get()})
            elif key == 'param':
                result['arguments'][key] = []
                for obj in value:
                    result['arguments'][key].append({'name': obj['name'].get(), 'value': obj['value'].get()})
            elif key == "hyperparameters":
                #result['hyperparameters'][key] = []
                for obj in value:
                    result['hyperparameters'].append({'name': obj['name'].get(), 'value': obj['value'].get()})
            elif key == "service_parameters":
                #result['service_parameters'][key] = []
                for obj in value:
                    result['service_parameters'].append({'name': obj['name'].get(), 'value': obj['value'].get()})
        result['steps'] = []
        for step in self._data[self._mode_sv.get()]["steps"]:
            step_result = {}
            #step_result['name'] = step['name'].get()
            #step_result['open'] = step['open']
            step_result['param'] = []
            for param in step['param']:
                # try converting the bounds to numbers
                #try:
                if param['type'].get() == 'float':
                    step_result['param'].append(
                        {
                            'name': param['name'].get(), 
                            'bounds': (float(param['bounds'][0].get()), 
                                       float(param['bounds'][1].get())),
                            'default_value': float(param['default_value'].get()),
                            'optimal_value': float(param['optimal_value'].get()),
                            'type': 'float',
                            'calibration_strategy': param['calibration_strategy'].get()
                        }
                    )
                elif param['type'].get() == 'list':
                    step_result['param'].append(
                        {
                            'name': param['name'].get(), 
                            'bounds': (float(param['bounds'][0].get()), 
                                       float(param['bounds'][1].get())),
                            'default_value': param['default_value'].get(),
                            'optimal_value': param['optimal_value'].get(),
                            'type': 'list',
                            'calibration_strategy': param['calibration_strategy'].get()
                        }
                    )
            step_result['objfunc'] = []
            for objfunc in step['objfunc']:
                step_result['objfunc'].append({'name': objfunc['name'].get(), 'of': objfunc['of'].get(), 'weight': float(objfunc['weight'].get()), 'data': (objfunc['data'][0].get(), objfunc['data'][1].get())})
            result['steps'].append(step_result)
        return result


from cosu import utils
from csip import Client
from typing import List, Dict, Tuple
import queue, os


def csip_worker(reqq: queue.Queue, thread_no: int, stop, full_trace,
                url, files, arg_params, conf: Dict, metainfo: Dict) -> None:
    async_call = conf.get('async_call', True)  # default is async
    save_resp = conf.get('save_response_to', None)  # save response, set it to a folder if responses should be saved.

    while not stop():
        try:
            (rnd, step, iteration, particle, x, step_param, calib_params, objfunc, resq) = reqq.get(True, 0.5)
            # print(thread_no, particle)

            c = Client(metainfo=metainfo)

            all_params = {}

            # static params (from args)
            for param in arg_params:
                c.add_data(param['name'], param['value'])

            # particle params  (generated from steps)
            # for i, value in enumerate(x):
            # for idx, value in enumerate(x[particle, :]):
            #     c.add_data(step_param_names[idx], value)
            values = x[particle]
            idx = 0
            for p in step_param:
                value_type = p.get('type', 'float')
                if 'float' == value_type:
                    c.add_data(p['name'], values[idx])
                    all_params[p['name']] = values[idx]
                    idx = idx + 1
                elif 'list' == value_type:
                    base_name = p['name']
                    calibration_strategy = p.get('calibration_strategy', 'mean')

                    if 'mean' == calibration_strategy:
                        mean = values[idx]
                        default_value = p['default_value']
                        value = [(1 + mean) * x for x in default_value]
                        c.add_data(base_name, value)
                        all_params[base_name] = value
                        idx = idx + 1
                    elif 'single' == calibration_strategy:
                        default_value = p['default_value']
                        sub_index = 0
                        value = default_value.copy()
                        while sub_index < len(default_value):
                            value[sub_index] = values[idx]
                            idx = idx + 1
                            sub_index = sub_index + 1
                        c.add_data(base_name, value)
                        all_params[base_name] = value

            # other, previously calibrated params (other steps)
            for name, value in calib_params.items():
                c.add_data(name, value)
                all_params[name] = value

            # objective function info
            for of in objfunc:
                c.add_cosu(of['name'], of['of'], of['data'])
                # c.add_data(of['name'], (of['data'][0], of['data'][1]))

            print('.', end='', flush=True)

            try:
                # print(c)
                if async_call:
                    res = c.execute_async(url, files=files, conf=conf)
                else:
                    res = c.execute(url, files=files, conf=conf)

                # run_string = 'r{}s{}i{}p{}.json'.format(rnd, step, iteration, particle)

                if save_resp:
                    res.save_to(os.path.join(save_resp, 'r{}s{}i{}p{}.json'.format(rnd, step, iteration, particle)))

                if res.is_failed():
                    print(u'F')
                    # print(res)
                    resq.put((particle, 100))
                else:
                    # print(res)
                    # print(u'\u2714', end='', flush=True)
                    print(u'O', end='', flush=True)
                    cost = utils.calc_cost(res, objfunc)
                    resq.put((particle, cost))
                    # break

                if full_trace is not None:
                    full_trace.append((all_params, cost))
            except Exception as e:
                print(res)
                print(e)

            reqq.task_done()
        except queue.Empty:
            continue



import os
from csip import Client


class SamplerTask:
    task_id: int
    parameters: dict[str, any]
    step_param: list[dict[str, any]]
    objectives: list[dict[str, any]]
    static_parameters: dict[str, any]
    url: str
    files: list[str]
    metainfo: dict[str, any]
    conf: dict[str, any]
    result: dict[str, any]

    def __init__(self, task_id: int, parameters: dict[str, any], step_param: list[dict[str, any]],
                 objectives: list[dict[str, any]], static_parameters: dict[str, any], url: str,
                 files: list[str] = None, metainfo: dict[str, any] = None, conf: dict[str, any] = None):
        self.task_id = task_id
        assert (parameters is not None and len(parameters) > 0)
        self.parameters = parameters
        assert (step_param is not None and len(step_param) > 0)
        self.step_param = step_param
        assert (objectives is not None and len(objectives) > 0)
        self.objectives = objectives
        self.static_parameters = static_parameters if static_parameters is not None else []
        assert (url is not None and len(url) > 0)
        self.url = url
        self.files = files if files is not None else []
        self.metainfo = metainfo
        self.conf = conf

    def create_request(self) -> Client:
        request: Client = Client(metainfo=self.metainfo)

        for key, value in self.static_parameters.items():
            request.add_data(key, value)

        self.add_parameters_to_request(request)

        for of in self.objectives:
            request.add_cosu(of['name'], of['of'], of['data'])

        return request

    def add_parameters_to_request(self, request: Client):
        for p in self.step_param:
            value_type = p.get('type', 'float')
            base_name = p['name']
            if 'float' == value_type:
                request.add_data(base_name, self.parameters[base_name])
            elif 'list' == value_type:
                calibration_strategy = p.get('calibration_strategy', 'mean')

                if 'mean' == calibration_strategy:
                    mean = self.parameters[base_name]
                    default_value = p['default_value']
                    value = [(1 + mean) * x for x in default_value]
                    request.add_data(base_name, value)
                elif 'single' == calibration_strategy:
                    default_value = p['default_value']
                    sub_index = 0
                    value = default_value.copy()
                    while sub_index < len(default_value):
                        value[sub_index] = self.parameters["{}_{}".format(base_name, sub_index)]
                        sub_index = sub_index + 1
                    request.add_data(base_name, value)

    def run_task(self) -> bool:
        self.result = {}
        request: Client = self.create_request()
        async_call: bool = self.conf.get('async_call', True) if self.conf is not None else True
        # save response, set it to a folder if responses should be saved.
        save_resp = self.conf.get('save_response_to', None) if self.conf is not None else None
        successful: bool = False

        response: Client = None
        try:
            if async_call:
                response = request.execute_async(self.url, files=self.files, conf=self.conf)
            else:
                response = request.execute(self.url, files=self.files, conf=self.conf)

            successful = response.is_finished()
            if not successful:
                print(response)

            if save_resp:
                response.save_to(os.path.join(save_resp, 'task_{}.json'.format(self.task_id)))

            objectives: list[dict[str, str]] = response.get_metainfo("cosu")
            for of in objectives:
                self.result[of["name"]] = of["value"]
        except Exception as ex:
            print(ex)
            print(response)
            successful = False

        return successful
